<?php
/* (c) Sam Biggins (sam@evoluted.net)
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */
namespace Deployer;

desc('Create if does not exist');
task('ci:database:setup', function () {
	$databaseName = getOrCreateDatabaseName();
	$projectName = getenv('CI_PROJECT_NAME');
})->onStage('cd');

desc('Creates the .env file setup for the environment, only runs for CD stage');
task('ci:setup:env', function () {
	writeln('<info>Writing .env file</info>');

	$db_name = getDatabaseName();
	$db_user = get('cd_database_user');
	$db_pass = get('cd_database_password');
	$db_host = get('cd_database_host');

	switch (get('application_type')) {
		case 'cake70':
		case 'cake71':
				$envContents = "DB_NAME='${db_name}'
DB_USER='${db_user}'
DB_PASS='${db_pass}'
DB_HOST='${db_host}'

ENVIRONMENT='PREVIEW'

PREVIEW_URL='" . getenv('CI_ENVIRONMENT_URL') . "'
";
			break;
		case 'laravel70':
		case 'laravel71':
			$envContents = file_get_contents(get('release_path') . '/.env.cd');

			// Replace database information
			$envContents = preg_replace("/DB_DATABASE(.*)/i", "DB_DATABASE='${db_name}'", $envContents);
			$envContents = preg_replace("/DB_HOST(.*)/i", "DB_HOST='${db_host}'", $envContents);
			$envContents = preg_replace("/DB_USERNAME(.*)/i", "DB_USERNAME='${db_user}'", $envContents);
			$envContents = preg_replace("/DB_PASSWORD(.*)/i", "DB_PASSWORD='${db_pass}'", $envContents);
			$envContents = preg_replace("/APPLICATION_DOMAIN(.*)/i", "APPLICATION_DOMAIN='" . getenv('CI_ENVIRONMENT_URL') . "'", $envContents);
			break;
		default:
			$envContents = '';
			break;
	}

	// Add .env file to deployment
	run('echo "' . $envContents . '" > ' . get('release_path') . "/.env");
})->onStage('cd');


desc('Rsync across assets');
task('ci:setup:assets', function () {

});

desc('Configure Vhost and PHP-FPM');
task('ci:vhost:configure', function () {
	$deploymentUrl = getenv('CI_ENVIRONMENT_URL');
	$hostname = preg_replace('#^https?://#', '', $deploymentUrl);

	run('vhost_gen.py -p {{release_path}} -t /etc/vhost-gen/custom-templates -n ' . $hostname . ' -c /etc/vhost-gen/' . get('application_type') . '.yml -s');
	run('sudo nginx -t && sudo service nginx reload');

	reloadWebserver();

})->onStage('cd');

desc('Creates a database for the deployment');
task('ci:database:create', function () {
	$databaseName = getDatabaseName();
	if (createDatabase($databaseName)) {
		// If this is not the staging branch, then we can
		// use the staging database to seed our new one if it exists
		writeln('<info>looking to seed the database</info>');
		if ($seedDatabaseName = getSeedDatabaseName()) {
			// Don't want to seed from the same database
			if ($seedDatabaseName != getDatabaseName()) {
				writeln('<info>Seeding from ' . $seedDatabaseName . '</info>');
				run("mysqldump --add-drop-table -u " . get('cd_database_user') . " -h " . get('cd_database_host') . " -p" . get('cd_database_password') . " " . $seedDatabaseName . " | mysql -u " . get('cd_database_user') . " -h " . get('cd_database_host') . " -p" . get('cd_database_password') . " " . getDatabaseName());
			}
		} else {
			writeln('<error>Cant find seed database</error>');
		}
	} else {
		writeln('<error>Could not create database.</error>');
	}
});

desc('Removes the environemnt from the server, used by Gitlab CD to tidy up');
task('ci:destroy', function () {
	// Get the database name to remove
	$databaseName = getDatabaseName();
	$deployPath = get('deploy_path');

	if (!preg_match('#^/var/www/cd/#', $deployPath)) {
		// Just a safety catch incase deploy path is set to / by accident or something
		// we would like to limit our damage
		return false;
	}

	run("rm -rf $deployPath");

	writeln('<error>WARNING: About to remove databse: ' . $databaseName . " and folder directory " . $deployPath);
	dropDatabase($databaseName);

	writeln('<error>WARNING: Removing vhost</error>');
	$hostname = preg_replace('#^https?://#', '', getenv('CI_ENVIRONMENT_URL'));
	run('rm /etc/nginx/conf.d/' . $hostname . '.conf');

	reloadWebserver();

});

/**
 * Forms the database name from environment variables, does not check for existance
 * @param  $branchName 	Allows overriding the branch name to find the db name for a different branch
 * @return string 		Database name
 */
function getDatabaseName($branchName = null) {
	$projectName = getenv('CI_PROJECT_NAME');

	if ($branchName === null) {
		$branchName = getenv('CI_COMMIT_REF_SLUG');
	}

	$databaseName = 'cd_' . $projectName . '_' . $branchName;

	// Swap - for _ to keep consistency
	$databaseName = str_replace('-', '_', $databaseName);

	// MySQL has a CHAR limit of 63 characters for database so if our database name is too long we
	// can shorten it using a hash while still trying to keep it useful
	$maxDatabaseNameLength = 63;
	if (strlen($databaseName) >= $maxDatabaseNameLength) {
		// CRC32 the database name and then convert to hexidecimal to shorten it, isn't fully collision
		// proof but will be for a string this long
		$databaseNameHash = base_convert(crc32($databaseName), 10, 32);
		$newDatabaseName = 'cd_' . $projectName . '_' . $databaseNameHash;

		// With our remaining chars use a substring of the branch name
		$subStringBranchName = substr($branchName, strlen($newDatabaseName) - $maxDatabaseNameLength);

		$newDatabaseName .= "_" . $subStringBranchName;

		// Swap - for _ to keep consistency
		$newDatabaseName = str_replace('-', '_', $newDatabaseName);

		$databaseName = $newDatabaseName;
	}

	return $databaseName;
}

/**
 * Returns the same of the seed database to copy initial data from
 */
function getSeedDatabaseName() {
	if (get('seed_database', false)) {
		$seedDatabase = get('seed_database');
	} else {
		// If we don't have a seed database explicitly set then we want
		// to use the staging database if there is one
		$seedDatabase = getDatabaseName('staging');
	}

	if (databaseExists($seedDatabase)) {
		return $seedDatabase;
	}
	// If seed database doesn't exist then we can't use it
	return false;
}

/**
 * Gets the database name if it exists otherwise it creates the database
 * @return string Database name
 */
function getOrCreateDatabaseName() {
	$databaseName = getDatabaseName();

	if (databaseExists($databaseName)) {
		writeln('<info>DATABASE does exist</info>');
	} else {
		writeln('<comment>Creating database: ' . $databaseName . '</comment>' );
		invoke('ci:database:create');
	}

	return $databaseName;
}

/**
 * Checks if a given database exists
 * @param  string $databaseName Name of database to check
 * @return boolean              True is the database exists
 */
function databaseExists($databaseName) {
	$databaseExists = false;
	$conn = new \mysqli(get('cd_database_host'), get('cd_database_user'), get('cd_database_password'));

	// Create database
	$sql = "SHOW DATABASES LIKE '$databaseName'";
	$result = $conn->query($sql);

	if ($result->num_rows > 0) {
		$databaseExists = true;
	}

	$conn->close();

	return $databaseExists;
}

/**
 * Creates a database of a given name
 * @param  string $databaseName Name of database to create
 * @return boolean              True is the database has been created
 */
function createDatabase($databaseName) {
	$databaseCreated = false;
	$conn = new \mysqli(get('cd_database_host'), get('cd_database_user'), get('cd_database_password'));

	// Create database
	$sql = "CREATE DATABASE $databaseName;";
	if ($conn->query($sql) === true) {
		$databaseCreated = true;
	}

	$conn->close();

	return $databaseCreated;
}

/**
 * Creates a database of a given name
 * @param  string $databaseName Name of database to create
 * @return boolean              True is the database has been created
 */
function dropDatabase($databaseName) {
	$databaseCreated = false;
	$conn = new \mysqli(get('cd_database_host'), get('cd_database_user'), get('cd_database_password'));

	$backupPath = '/var/www/cd/backups/mysql/' . getenv('CI_PROJECT_NAME');
	if (!file_exists($backupPath)) {
		mkdir($backupPath);
	}

	$dumpLocation = '/var/www/cd/backups/mysql/' . getenv('CI_PROJECT_NAME') . '/' . getenv('CI_COMMIT_REF_SLUG') . '.sql';

	// Before we drop the database lets take a backup and file it away just incase we need to restore it
	dumpDatabase($databaseName, $dumpLocation);

	// Drop database
	$sql = "DROP DATABASE IF EXISTS $databaseName;";
	if ($conn->query($sql) === true) {
		$databaseCreated = true;
	}

	$conn->close();

	return $databaseCreated;
}

/**
 * Dump given database into given file location
 * @param  string $databaseName Database to dump/backup
 * @param  string $location 	File location to dump file to
 */
function dumpDatabase($databaseName, $dumpLocation) {
	run("mysqldump --add-drop-table -u " . get('cd_database_user') . " -h " . get('cd_database_host') . " -p" . get('cd_database_password') . " " . $databaseName . " > " . $dumpLocation);
}

/**
 * Reloads the webserver, handles different applications webserver types
 */
function reloadWebserver() {
	// Based on the application type there may be different services to reload
	switch (get('application_type')) {
		case 'laravel70':
		case 'laravel71':
		case 'cake70':
		case 'cake71':
			run('sudo service php7.0-fpm reload');
			break;
	}
}

